from __future__ import unicode_literals

from cytoolz.functoolz import (
    compose,
    partial,
)

from eth_utils import (
    int_to_big_endian,
    pad_left,
)


zpad = partial(pad_left, pad_with=b'\x00')
zpad32 = partial(pad_left, to_size=32, pad_with=b'\x00')


int_to_32byte_big_endian = compose(
    zpad32,
    int_to_big_endian,
)



import functools
import time

from cytoolz.dicttoolz import (
    assoc,
)

from eth_utils import (
    keccak,
    to_dict,
    is_dict,
    is_integer,
    is_list_like,
    is_text,
    is_bytes,
    is_null,
    force_bytes,
    to_tuple,
    apply_to_return_value,
)

from eth_tester.utils.accounts import (
    generate_contract_address,
)


ZERO_32BYTES = b'\x00' * 32
ZERO_8BYTES = b'\x00' * 8
ZERO_ADDRESS = b'\x00' * 20


@apply_to_return_value(b'|'.join)
@to_tuple
def stringify(value):
    if is_bytes(value):
        yield value
    elif is_text(value):
        yield force_bytes(value)
    elif is_list_like(value):
        yield b''.join((
            b'(',
            b','.join((stringify(item) for item in value)),
            b')',
        ))
    elif is_dict(value):
        yield b''.join((
            b'{',
            b','.join((
                b":".join((stringify(key), stringify(item)))
                for key, item
                in value.items()
            )),
            b'}',
        ))
    elif is_integer(value):
        yield force_bytes(str(value))
    elif is_null(value):
        yield 'None@{0}'.format(id(value))
    else:
        raise TypeError("Unsupported type for stringification: {0}".format(type(value)))


def fake_rlp_hash(value):
    return keccak(stringify(value))


def add_hash(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        value = fn(*args, **kwargs)
        if 'hash' in value:
            return value
        else:
            return assoc(value, 'hash', keccak(stringify(value)))
    return inner


@to_dict
def create_transaction(transaction, block, transaction_index, is_pending, overrides=None):
    if overrides is None:
        overrides = {}

    if 'nonce' in overrides:
        yield 'nonce', overrides['nonce']
    else:
        yield 'nonce', 0

    if 'hash' in overrides:
        yield 'hash', overrides['hash']
    else:
        yield 'hash', fake_rlp_hash(transaction)

    if 'from' in overrides:
        yield 'from', overrides['from']
    else:
        yield 'from', transaction['from']

    if 'gas' in overrides:
        yield 'gas', overrides['gas']
    else:
        yield 'gas', transaction['gas']

    if 'gas_price' in overrides:
        yield 'gas_price', overrides['gas_price']
    else:
        yield 'gas_price', transaction.get('gas_price', 1)  # TODO: make configurable

    if 'to' in overrides:
        yield 'to', overrides['to']
    else:
        yield 'to', transaction.get('to', b'')

    if 'data' in overrides:
        yield 'data', overrides['data']
    else:
        yield 'data', transaction.get('data', b'')

    if 'value' in overrides:
        yield 'value', overrides['value']
    else:
        yield 'value', transaction.get('value', 0)

    if 'nonce' in overrides:
        yield 'nonce', overrides['nonce']
    else:
        yield 'nonce', transaction.get('nonce', 0)

    if 'v' in overrides:
        yield 'v', overrides['v']
    else:
        yield 'v', transaction.get('v', 27)

    if 'r' in overrides:
        yield 'r', overrides['r']
    else:
        yield 'r', transaction.get('r', 12345)

    if 's' in overrides:
        yield 's', overrides['s']
    else:
        yield 's', transaction.get('s', 67890)


@to_dict
def make_log(transaction, block, transaction_index, log_index, overrides=None):
    if overrides is None:
        overrides = {}

    is_pending = transaction['block_number'] is None

    if 'type' in overrides:
        yield 'type', overrides['type']
    else:
        yield 'type', 'pending' if is_pending else 'mined'

    if 'transaction_index' in overrides:
        yield 'transaction_index', overrides['transaction_index']
    else:
        yield 'transaction_index', None if is_pending else transaction_index

    if 'block_number' in overrides:
        yield 'block_number', overrides['block_number']
    else:
        yield 'block_number', None if is_pending else block['number']

    if 'block_hash' in overrides:
        yield 'block_hash', overrides['block_hash']
    else:
        yield 'block_hash', None if is_pending else block['hash']

    if 'log_index' in overrides:
        yield 'log_index', overrides['log_index']
    else:
        yield 'log_index', log_index

    if 'address' in overrides:
        yield 'address', overrides['address']
    else:
        yield 'address', transaction.get('to', b'')

    if 'data' in overrides:
        yield 'data', overrides['data']
    else:
        yield 'data', b''

    if 'topics' in overrides:
        yield 'topics', overrides['topics']
    else:
        yield 'topics', []


@to_dict
def make_receipt(transaction, block, transaction_index, overrides=None):
    if overrides is None:
        overrides = {}

    if 'gas_used' in overrides:
        gas_used = overrides['gas_used']
    else:
        gas_used = 21000
    yield 'gas_used', gas_used

    if 'cumulative_gas_used' in overrides:
        yield 'cumulative_gas_used', overrides['cumulative_gas_used']
    else:
        yield 'cumulative_gas_used', block['gas_used'] + gas_used

    if 'contract_address' in overrides:
        yield 'contract_address', overrides['contract_address']
    else:
        contract_address = generate_contract_address(transaction['from'], transaction['nonce'])
        yield 'contract_address', contract_address

    if 'logs' in overrides:
        yield 'logs', overrides['logs']
    else:
        yield 'logs', []

    if 'transaction_hash' in overrides:
        yield 'transaction_hash', overrides['transaction_hash']
    else:
        yield 'transaction_hash', transaction['hash']


GENESIS_NONCE = b'\x00\x00\x00\x00\x00\x00\x00*'  # 42 encoded as big-endian-integer
BLANK_ROOT_HASH = b'V\xe8\x1f\x17\x1b\xccU\xa6\xff\x83E\xe6\x92\xc0\xf8n\x5bH\xe0\x1b\x99l\xad\xc0\x01b/\xb5\xe3c\xb4!'  # noqa: E501
EMPTY_UNCLE_HASH = b'\x1d\xccM\xe8\xde\xc7]z\xab\x85\xb5g\xb6\xcc\xd4\x1a\xd3\x12E\x1b\x94\x8at\x13\xf0\xa1B\xfd@\xd4\x93G'  # noqa: E501


def make_genesis_block():
    return {
        "number": 0,
        "hash": ZERO_32BYTES,
        "parent_hash": ZERO_32BYTES,
        "nonce": GENESIS_NONCE,
        "sha3_uncles": EMPTY_UNCLE_HASH,
        "logs_bloom": 0,
        "transactions_root": BLANK_ROOT_HASH,
        "receipts_root": BLANK_ROOT_HASH,
        "state_root": BLANK_ROOT_HASH,
        "miner": ZERO_ADDRESS,
        "difficulty": 131072,
        "total_difficulty": 131072,
        "size": 0,
        "extra_data": ZERO_32BYTES,
        "gas_limit": 3141592,
        "gas_used": 0,
        "timestamp": int(time.time()),
        "transactions": [],
        "uncles": [],
    }


@add_hash
@to_dict
def make_block_from_parent(parent_block, overrides=None):
    if overrides is None:
        overrides = {}

    if 'number' in overrides:
        yield 'number', overrides['number']
    else:
        yield 'number', parent_block['number'] + 1

    if 'hash' in overrides:
        yield 'hash', overrides['hash']
    else:
        yield 'hash', keccak(parent_block['hash'])

    if 'parent_hash' in overrides:
        yield 'parent_hash', overrides['parent_hash']
    else:
        yield 'parent_hash', parent_block['hash']

    if 'nonce' in overrides:
        yield 'nonce', overrides['nonce']
    else:
        yield 'nonce', parent_block['nonce']

    if 'sha3_uncles' in overrides:
        yield 'sha3_uncles', overrides['sha3_uncles']
    else:
        yield 'sha3_uncles', EMPTY_UNCLE_HASH

    if 'logs_bloom' in overrides:
        yield 'logs_bloom', overrides['logs_bloom']
    else:
        yield 'logs_bloom', 0

    if 'transactions_root' in overrides:
        yield 'transactions_root', overrides['transactions_root']
    else:
        yield 'transactions_root', BLANK_ROOT_HASH

    if 'receipts_root' in overrides:
        yield 'receipts_root', overrides['receipts_root']
    else:
        yield 'receipts_root', BLANK_ROOT_HASH

    if 'state_root' in overrides:
        yield 'state_root', overrides['state_root']
    else:
        yield 'state_root', BLANK_ROOT_HASH

    if 'miner' in overrides:
        yield 'miner', overrides['miner']
    else:
        yield 'miner', ZERO_ADDRESS

    if 'difficulty' in overrides:
        difficulty = overrides['difficulty']
    else:
        difficulty = 131072
    yield 'difficulty', difficulty

    if 'total_difficulty' in overrides:
        yield 'total_difficulty', overrides['total_difficulty']
    else:
        yield 'total_difficulty', parent_block['difficulty'] + difficulty

    if 'size' in overrides:
        yield 'size', overrides['size']
    else:
        yield 'size', 0

    if 'extra_data' in overrides:
        yield 'extra_data', overrides['extra_data']
    else:
        yield 'extra_data', ZERO_32BYTES

    if 'gas_limit' in overrides:
        yield 'gas_limit', overrides['gas_limit']
    else:
        yield 'gas_limit', parent_block['gas_limit']

    if 'gas_used' in overrides:
        yield 'gas_used', overrides['gas_used']
    else:
        yield 'gas_used', 0

    if 'timestamp' in overrides:
        yield 'timestamp', overrides['timestamp']
    else:
        yield 'timestamp', parent_block['timestamp'] + 15

    if 'transactions' in overrides:
        yield 'transactions', overrides['transactions']
    else:
        yield 'transactions', []

    if 'uncles' in overrides:
        yield 'uncles', overrides['uncles']
    else:
        yield 'uncles', []



from eth_utils import (
    is_address,
)

from eth_tester.exceptions import (
    ValidationError,
)


TRANSACTION_KEYS = {
    'from',
    'to',
    'gas',
    'gas_price',
    'value',
    'data',
}


def validate_transaction(transaction):
    # TODO: refactor this so that it's not gross.
    if 'from' not in transaction:
        raise ValidationError("Transactions must specify a 'from' address")
    elif not is_address(transaction['from']):
        raise ValidationError("transaction[from]: Unrecognized address format: {0}".format(
            transaction['from'],
        ))
    elif 'to' in transaction and not is_address(transaction['to']):
        raise ValidationError("transaction[to]: Unrecognized address format: {0}".format(
            transaction['to'],
        ))

    extra_keys = set(transaction.keys()).difference(TRANSACTION_KEYS)
    if extra_keys:
        raise ValidationError(
            "Transactions may only include the keys {0}.  The following extra "
            "keys were found: {1}".format(
                ",".join(sorted(TRANSACTION_KEYS)),
                ",".join(sorted(extra_keys)),
            )
        )



"""
Functions lifted from https://github.com/vbuterin/pybitcointools
"""
from eth_utils import (
    big_endian_to_int,
    int_to_big_endian,
    is_bytes,
    pad_left,
)

from eth_tester.constants import (
    SECPK1_G,
    SECPK1_N,
)

from .jacobian import (
    fast_multiply,
)


def _pad32(value):
    return pad_left(value, 32, b'\x00')


def _encode_raw_public_key(raw_public_key):
    left, right = raw_public_key
    return b''.join((
        _pad32(int_to_big_endian(left)),
        _pad32(int_to_big_endian(right)),
    ))


def private_key_to_public_key(private_key):
    if not is_bytes(private_key) or len(private_key) != 32:
        raise TypeError("`private_key` must be of type `bytes` and of lenght 32")
    private_key_as_num = big_endian_to_int(private_key)

    if private_key_as_num >= SECPK1_N:
        raise Exception("Invalid privkey")

    raw_public_key = fast_multiply(SECPK1_G, private_key_as_num)
    public_key = _encode_raw_public_key(raw_public_key)
    return public_key



from __future__ import absolute_import

import pkg_resources

from semantic_version import (
    Spec,
)

from cytoolz.dicttoolz import (
    assoc,
)

import rlp

from eth_utils import (
    encode_hex,
    remove_0x_prefix,
    to_tuple,
)

from eth_tester.constants import (
    FORK_HOMESTEAD,
    FORK_DAO,
    FORK_ANTI_DOS,
    FORK_STATE_CLEANUP,
)
from eth_tester.exceptions import (
    BlockNotFound,
    TransactionNotFound,
    UnknownFork,
    TransactionFailed,
)
from eth_tester.backends.base import BaseChainBackend
from eth_tester.backends.pyethereum.utils import (
    get_pyethereum_version,
    is_pyethereum16_available,
)

from eth_tester.utils.accounts import (
    private_key_to_address,
)

from eth_tester.backends.pyethereum.serializers import (
    serialize_block,
    serialize_transaction,
    serialize_transaction_hash,
    serialize_transaction_receipt,
)
from eth_tester.backends.pyethereum.validation import (
    validate_transaction,
)
from eth_tester.utils.formatting import (
    replace_exceptions,
)


if is_pyethereum16_available():
    from ethereum.tester import (
        TransactionFailed as Pyeth16TransactionFailed,
    )
else:
    Pyeth16TransactionFailed = None


#
# Internal getters for EVM objects
#
def _get_transaction_by_hash(evm, transaction_hash, mined=True):
    # first check unmined transactions
    for index, candidate in enumerate(evm.block.get_transaction_hashes()):
        if candidate == transaction_hash:
            transaction = evm.block.transaction_list[index]
            return evm.block, transaction, index

    # then check work backwards through the blocks looking for mined transactions.
    for block in reversed(evm.blocks[:-1]):
        for index, candidate in enumerate(block.get_transaction_hashes()):
            if candidate == transaction_hash:
                transaction = block.transaction_list[index]
                return block, transaction, index
    else:
        raise TransactionNotFound(
            "No transaction found for transaction hash {0}".format(
                encode_hex(transaction_hash),
            )
        )


def _get_block_by_number(evm, block_number="latest"):
    if block_number == "latest":
        if len(evm.blocks) == 0:
            raise BlockNotFound("Chain has no blocks")
        elif len(evm.blocks) == 1:
            return evm.blocks[0]
        else:
            return evm.blocks[-2]
    elif block_number == "earliest":
        return evm.blocks[0]
    elif block_number == "pending":
        return evm.block
    elif block_number == evm.block.number:
        return evm.block
    else:
        if block_number >= len(evm.blocks):
            raise BlockNotFound("Block number is longer than current chain.")
        return evm.blocks[block_number]


def _get_block_by_hash(evm, block_hash):
    block_hash = block_hash

    for block in reversed(evm.blocks):
        if block.hash == block_hash:
            return block
    else:
        if block_hash == evm.block.hash:
            return evm.block
        raise BlockNotFound("Could not find block for provided hash")


def _send_evm_transaction(tester_module, evm, transaction):
    from ethereum import tester

    try:
        # record the current gas price so that it can be reset after sending
        # the transaction.
        pre_transaction_gas_price = tester.gas_price
        pre_transaction_gas_limit = tester.gas_limit
        # set the evm gas price to the one specified by the transaction.
        tester.gas_price = transaction.get('gas_price', tester.gas_price)
        tester.gas_limit = transaction['gas']

        # get the private key of the sender.
        try:
            sender = tester.keys[tester.accounts.index(transaction['from'])]
        except ValueError:
            sender = evm.extra_accounts[transaction['from']]

        output = evm.send(
            sender=sender,
            to=transaction.get('to', b''),
            value=transaction.get('value', 0),
            evmdata=transaction.get('data', b''),
        )
    finally:
        # revert the tester gas price back to the original value.
        tester.gas_price = pre_transaction_gas_price
        tester.gas_limit = pre_transaction_gas_limit

    return output


def _estimate_evm_transaction(tester_module, evm, transaction):
    transaction_for_estimate = assoc(transaction, 'gas', 50000000)
    return _send_evm_transaction(tester_module, evm, transaction_for_estimate)


def _call_evm_transaction(tester_module, evm, transaction):
    if 'gas' not in transaction:
        transaction_for_call = assoc(transaction, 'gas', 50000000)
    else:
        transaction_for_call = transaction
    return _send_evm_transaction(tester_module, evm, transaction_for_call)


class PyEthereum16Backend(BaseChainBackend):
    def __init__(self):
        if not is_pyethereum16_available():
            version = get_pyethereum_version()
            if version is None:
                raise pkg_resources.DistributionNotFound(
                    "The `ethereum` package is not available.  The "
                    "`PyEthereum16Backend` requires a 1.6.x version of the "
                    "ethereum package to be installed."
                )
            elif version not in Spec('>=1.6.0,<1.7.0'):
                raise pkg_resources.DistributionNotFound(
                    "The `PyEthereum16Backend` requires a 1.6.x version of the "
                    "`ethereum` package.  Found {0}".format(version)
                )
        self.reset_to_genesis()

    #
    # Fork Rules
    #
    def set_fork_block(self, fork_name, fork_block):
        if fork_name == FORK_HOMESTEAD:
            self.evm.env.config['HOMESTEAD_FORK_BLKNUM'] = fork_block
        elif fork_name == FORK_DAO:
            self.evm.env.config['DAO_FORK_BLKNUM'] = fork_block
        elif fork_name == FORK_ANTI_DOS:
            self.evm.env.config['ANTI_DOS_FORK_BLKNUM'] = fork_block
        elif fork_name == FORK_STATE_CLEANUP:
            self.evm.env.config['CLEARING_FORK_BLKNUM'] = fork_block
        else:
            raise UnknownFork("Unknown fork name: {0}".format(fork_name))

    def get_fork_block(self, fork_name):
        if fork_name == FORK_HOMESTEAD:
            return self.evm.env.config['HOMESTEAD_FORK_BLKNUM']
        elif fork_name == FORK_DAO:
            return self.evm.env.config['DAO_FORK_BLKNUM']
        elif fork_name == FORK_ANTI_DOS:
            return self.evm.env.config['ANTI_DOS_FORK_BLKNUM']
        elif fork_name == FORK_STATE_CLEANUP:
            return self.evm.env.config['CLEARING_FORK_BLKNUM']
        else:
            raise UnknownFork("Unknown fork name: {0}".format(fork_name))

    #
    # Snapshots
    #
    def take_snapshot(self):
        return (self.evm.block.number, self.evm.snapshot())

    def revert_to_snapshot(self, snapshot):
        from ethereum import tester

        block_number, snapshot_data = snapshot

        # Remove all blocks after our saved block number.
        del self.evm.blocks[block_number:]

        self.evm.revert(snapshot_data)

        if self.evm.blocks:
            parent = self.evm.blocks[-1]
            block = self.evm.block.init_from_parent(
                parent,
                tester.DEFAULT_ACCOUNT,
                timestamp=parent.timestamp + 6,
            )

            self.evm.block = block
            self.evm.blocks.append(block)
        else:
            self.evm.blocks.append(self.evm.block)

    def reset_to_genesis(self):
        from ethereum import tester
        self.evm = tester.state()
        self.evm.extra_accounts = {}
        self.mine_blocks()

    #
    # Meta
    #
    def time_travel(self, to_timestamp):
        from ethereum import tester

        self.evm.block.finalize()
        self.evm.block.commit_state()
        self.evm.db.put(
            self.evm.block.hash,
            rlp.encode(self.evm.block),
        )

        block = self.evm.block.init_from_parent(
            self.evm.block,
            tester.DEFAULT_ACCOUNT,
            timestamp=to_timestamp,
        )

        self.evm.block = block
        self.evm.blocks.append(block)
        return to_timestamp

    #
    # Mining
    #
    @to_tuple
    def mine_blocks(self, num_blocks=1, coinbase=None):
        from ethereum import tester

        if coinbase is None:
            coinbase = tester.DEFAULT_ACCOUNT

        self.evm.mine(
            number_of_blocks=num_blocks,
            coinbase=coinbase,
        )
        for block in self.evm.blocks[-1 * num_blocks - 1:-1]:
            yield block.hash

    #
    # Accounts
    #
    @to_tuple
    def get_accounts(self):
        from ethereum import tester

        for account in tester.accounts:
            yield account

        for account in self.evm.extra_accounts.keys():
            yield account

    def add_account(self, private_key):
        account = private_key_to_address(private_key)
        self.evm.extra_accounts[account] = private_key

    #
    # Chain data
    #
    def get_block_by_number(self, block_number, full_transactions=False):
        if full_transactions:
            transaction_serialize_fn = serialize_transaction
        else:
            transaction_serialize_fn = serialize_transaction_hash

        block = _get_block_by_number(
            self.evm,
            block_number,
        )
        is_pending = block == self.evm.block
        return serialize_block(self.evm, block, transaction_serialize_fn, is_pending)

    def get_block_by_hash(self, block_hash, full_transactions=False):
        if full_transactions:
            transaction_serialize_fn = serialize_transaction
        else:
            transaction_serialize_fn = serialize_transaction_hash

        block = _get_block_by_hash(
            self.evm,
            block_hash,
        )
        is_pending = block == self.evm.block
        return serialize_block(self.evm, block, transaction_serialize_fn, is_pending)

    def get_latest_block(self, full_transactions=False):
        if full_transactions:
            transaction_serialize_fn = serialize_transaction
        else:
            transaction_serialize_fn = serialize_transaction_hash

        return serialize_block(self.evm, self.evm.block, transaction_serialize_fn)

    def get_transaction_by_hash(self, transaction_hash):
        block, transaction, transaction_index = _get_transaction_by_hash(
            self.evm,
            transaction_hash,
        )
        is_pending = block.number == self.evm.block.number
        return serialize_transaction(block, transaction, transaction_index, is_pending)

    def get_transaction_receipt(self, transaction_hash):
        block, transaction, transaction_index = _get_transaction_by_hash(
            self.evm,
            transaction_hash,
        )
        is_pending = block.number == self.evm.block.number
        return serialize_transaction_receipt(
            block,
            transaction,
            block.get_receipt(transaction_index),
            transaction_index,
            is_pending,
        )

    #
    # Account state
    #
    def get_nonce(self, account, block_number="latest"):
        block = _get_block_by_number(self.evm, block_number)
        return block.get_nonce(remove_0x_prefix(account))

    def get_balance(self, account, block_number="latest"):
        block = _get_block_by_number(self.evm, block_number)
        return block.get_balance(remove_0x_prefix(account))

    def get_code(self, account, block_number="latest"):
        block = _get_block_by_number(self.evm, block_number)
        return block.get_code(remove_0x_prefix(account))

    #
    # Transactions
    #
    def send_raw_transaction(self, raw_transaction):
        from ethereum import (
            processblock,
            tester,
        )
        from ethereum.transactions import Transaction
        rlp_transaction = rlp.decode(raw_transaction, Transaction)
        # Manipulate `tester.state` directly
        success, _ = processblock.apply_transaction(self.evm.block, rlp_transaction)
        if not success:
            raise tester.TransactionFailed
        self.evm.last_tx = rlp_transaction
        return rlp_transaction.hash

    def send_transaction(self, transaction):
        from ethereum import tester
        validate_transaction(transaction)
        _send_evm_transaction(
            tester_module=tester,
            evm=self.evm,
            transaction=transaction,
        )
        return self.evm.last_tx.hash

    @replace_exceptions({Pyeth16TransactionFailed: TransactionFailed})
    def call(self, transaction, block_number="latest"):
        from ethereum import tester
        validate_transaction(transaction)

        if block_number != "latest":
            raise NotImplementedError("Block number must be 'latest'.")

        snapshot = self.take_snapshot()
        try:
            output = _call_evm_transaction(
                tester_module=tester,
                evm=self.evm,
                transaction=transaction,
            )
        except TransactionFailed as e:
            raise e
        finally:
            self.revert_to_snapshot(snapshot)
        return output

    @replace_exceptions({Pyeth16TransactionFailed: TransactionFailed})
    def estimate_gas(self, transaction):
        from ethereum import tester
        validate_transaction(transaction)

        snapshot = self.take_snapshot()
        try:
            _estimate_evm_transaction(
                tester_module=tester,
                evm=self.evm,
                transaction=transaction,
            )
            txn_hash = self.evm.last_tx.hash
            receipt = self.get_transaction_receipt(txn_hash)
        except TransactionFailed as e:
            raise e
        finally:
            self.revert_to_snapshot(snapshot)
        return receipt['gas_used']



from __future__ import unicode_literals

from eth_utils import (
    encode_hex,
    function_abi_to_4byte_selector,
)


THROWS_SOURCE = (
    """
    contract Test {
        uint public value = 1;
        function setValue(uint _value) {
            value = _value;
        }
        function willThrow() {
            require(false);
        }
    }
    """
)


THROWS_BYTECODE = (
    "60606040526001600055341561001457600080fd5b60f2806100226000396000f3006060604052600"
    "0357c0100000000000000000000000000000000000000000000000000000000900463ffffffff1680"
    "6318955b1e1460505780633fa4f2451460625780635524107714608857600080fd5b3415605a57600"
    "080fd5b606060a8565b005b3415606c57600080fd5b607260b6565b60405180828152602001915050"
    "60405180910390f35b3415609257600080fd5b60a6600480803590602001909190505060bc565b005"
    "b6000151560b457600080fd5b565b60005481565b80600081905550505600a165627a7a72305820d5"
    "8a6b595eeb1a765e71924eec6e96c98c1fe9e90876ec232593b9ebb9c686500029"
)


THROWS_ABI = {
    'willThrow': {
        'constant': False,
        'inputs': [],
        'name': 'willThrow',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'value': {
        'constant': True,
        'inputs': [],
        'name': 'value',
        'outputs': [{'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'setValue': {
        'constant': False,
        'inputs': [{'name': '_value', 'type': 'uint256'}],
        'name': 'setValue',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
}


def _deploy_throws(eth_tester):
    deploy_hash = eth_tester.send_transaction({
        "from": eth_tester.get_accounts()[0],
        "gas": 500000,
        "data": THROWS_BYTECODE,
    })
    deploy_receipt = eth_tester.get_transaction_receipt(deploy_hash)
    throws_address = deploy_receipt['contract_address']
    assert throws_address
    throws_code = eth_tester.get_code(throws_address)
    assert len(throws_code) > 2
    return throws_address


def _make_call_throws_transaction(eth_tester, contract_address, fn_name, fn_args=None):
    from eth_abi import encode_abi

    if fn_args is None:
        fn_args = tuple()

    fn_abi = THROWS_ABI[fn_name]
    arg_types = [
        arg_abi['type']
        for arg_abi
        in fn_abi['inputs']
    ]
    fn_selector = function_abi_to_4byte_selector(fn_abi)
    transaction = {
        "from": eth_tester.get_accounts()[0],
        "to": contract_address,
        "gas": 500000,
        "data": encode_hex(fn_selector + encode_abi(arg_types, fn_args)),
    }
    return transaction


def _decode_throws_result(fn_name, result):
    from eth_abi import decode_abi

    fn_abi = THROWS_ABI[fn_name]
    output_types = [
        output_abi['type']
        for output_abi
        in fn_abi['outputs']
    ]

    return decode_abi(output_types, result)



from __future__ import absolute_import

import functools

from cytoolz.functoolz import (
    curry,
)

from eth_utils import (
    to_dict,
    to_list,
)


def hex_to_integer(value):
    return int(value, 16)


@curry
@to_list
def apply_formatter_at_index(formatter, at_index, value):
    if at_index + 1 > len(value):
        raise IndexError(
            "Not enough values in iterable to apply formatter.  Got: {0}. "
            "Need: {1}".format(len(value), at_index)
        )
    for index, item in enumerate(value):
        if index == at_index:
            yield formatter(item)
        else:
            yield item


@curry
def apply_formatter_if(formatter, condition, value):
    if condition(value):
        return formatter(value)
    else:
        return value


@curry
@to_dict
def apply_formatters_to_dict(formatters, value):
    for key, item in value.items():
        if key in formatters:
            yield key, formatters[key](item)
        else:
            yield key, item


@curry
@to_list
def apply_formatter_to_array(formatter, value):
    for item in value:
        yield formatter(item)


@curry
def apply_one_of_formatters(formatter_condition_pairs, value):
    for formatter, condition in formatter_condition_pairs:
        if condition(value):
            return formatter(value)
    else:
        raise ValueError("The provided value did not satisfy any of the formatter conditions")


@curry
@to_dict
def apply_key_map(key_mappings, value):
    for key, item in value.items():
        if key in key_mappings:
            yield key_mappings[key], item
        else:
            yield key, item


def replace_exceptions(old_to_new_exceptions):
    old_exceptions = tuple(old_to_new_exceptions.keys())

    def decorator(to_wrap):
        @functools.wraps(to_wrap)
        def wrapper(*args, **kwargs):
            try:
                return to_wrap(*args, **kwargs)
            except old_exceptions as e:
                try:
                    raise old_to_new_exceptions[type(e)] from e
                except KeyError:
                    raise TypeError("could not look up new exception to use for %r" % e) from e
        return wrapper
    return decorator



from __future__ import unicode_literals

import functools

from eth_utils import (
    is_boolean,
    is_bytes,
    is_text,
    is_dict,
    is_integer,
    is_list_like,
    to_dict,
    to_tuple,
)

from eth_tester.constants import (
    UINT256_MAX,
)
from eth_tester.exceptions import (
    ValidationError,
)


def validate_positive_integer(value):
    error_message = "Value must be positive integers.  Got: {0}".format(
        value,
    )
    if not is_integer(value) or is_boolean(value):
        raise ValidationError(error_message)
    elif value < 0:
        raise ValidationError(error_message)


def validate_uint256(value):
    validate_positive_integer(value)
    if value > UINT256_MAX:
        raise ValidationError("Value exceeds maximum 256 bit integer size:  {0}".format(value))


def validate_bytes(value):
    if not is_bytes(value):
        raise ValidationError("Value must be a byte string.  Got type: {0}".format(type(value)))


def validate_text(value):
    if not is_text(value):
        raise ValidationError("Value must be a text string.  Got type: {0}".format(type(value)))


def validate_is_dict(value):
    if not is_dict(value):
        raise ValidationError("Value must be a dictionary.  Got: {0}".format(type(value)))


def validate_is_list_like(value):
    if not is_list_like(value):
        raise ValidationError("Value must be a sequence type.  Got: {0}".format(type(value)))


@to_tuple
def _accumulate_errors(value, validators):
    for idx, validator in enumerate(validators):
        try:
            validator(value)
        except ValidationError as err:
            yield idx, err


def validate_any(value, validators):
    errors = _accumulate_errors(value, validators)
    if len(errors) == len(validators):
        item_error_messages = tuple(
            " - [{0}]: {1}".format(idx, str(err))
            for idx, err
            in errors
        )
        error_message = (
            "Value did not pass any of the provided validators:\n"
            "{0}".format(
                "\n".join(item_error_messages)
            )
        )
        raise ValidationError(error_message)


def validate_no_extra_keys(value, allowed_keys):
    extra_keys = tuple(sorted(set(value.keys()).difference(allowed_keys)))
    if extra_keys:
        raise ValidationError(
            "Only the keys '{0}' are allowed.  Got extra keys: '{1}'".format(
                "/".join(tuple(sorted(allowed_keys))),
                "/".join(extra_keys),
            )
        )


def validate_has_required_keys(value, required_keys):
    missing_keys = tuple(sorted(set(required_keys).difference(value.keys())))
    if missing_keys:
        raise ValidationError(
            "Blocks must contain all of the keys '{0}'.  Missing the keys: '{1}'".format(
                "/".join(tuple(sorted(required_keys))),
                "/".join(missing_keys),
            )
        )


@to_dict
def _accumulate_dict_errors(value, validators):
    for key, validator_fn in validators.items():
        item = value[key]
        try:
            validator_fn(item)
        except ValidationError as err:
            yield key, err


def validate_dict(value, key_validators):
    validate_is_dict(value)
    validate_no_extra_keys(value, key_validators.keys())
    validate_has_required_keys(value, key_validators.keys())

    key_errors = _accumulate_dict_errors(value, key_validators)
    if key_errors:
        key_messages = tuple(
            "{0}: {1}".format(key, str(err))
            for key, err
            in sorted(key_errors.items())
        )
        error_message = (
            "The following keys failed to validate\n"
            "- {0}".format(
                "\n - ".join(key_messages)
            )
        )
        raise ValidationError(error_message)


@to_tuple
def _accumulate_array_errors(value, validator):
    for index, item in enumerate(value):
        try:
            validator(item)
        except ValidationError as err:
            yield index, err


def validate_array(value, validator):
    validate_is_list_like(value)

    item_errors = _accumulate_array_errors(value, validator)
    if item_errors:
        item_messages = tuple(
            "[{0}]: {1}".format(index, str(err))
            for index, err
            in sorted(item_errors)
        )
        error_message = (
            "The following items failed to validate\n"
            "- {0}".format(
                "\n - ".join(item_messages)
            )
        )
        raise ValidationError(error_message)


def if_not_null(validator_fn):
    @functools.wraps(validator_fn)
    def inner(value):
        if value is not None:
            validator_fn(value)
    return inner


def if_not_create_address(validator_fn):
    @functools.wraps(validator_fn)
    def inner(value):
        if value != b'':
            validator_fn(value)
    return inner



from cytoolz.functoolz import (
    curry,
)

from eth_utils import (
    to_dict,
    to_tuple,
)


@curry
@to_dict
def normalize_dict(value, normalizers):
    for key, item in value.items():
        normalizer = normalizers[key]
        yield key, normalizer(item)


@curry
@to_tuple
def normalize_array(value, normalizer):
    """
    This is just `map` but it's nice to have it return a consisten type
    (tuple).
    """
    for item in value:
        yield normalizer(item)


@curry
def normalize_if(value, conditional_fn, normalizer):
    if conditional_fn(value):
        return normalizer(value)
    else:
        return value
